/*
	Author: B.McA (jdoe407)
	Date: March 05, 2005

	Purpose: Generic binary patcher class
	Usage: filePatcher varName( executable, array of patch struct's, # of patches in patch array, output function );
	Todo: Take advantage of the class structure a bit more other then a constructor/destructor
*/

#include "patcher.h"
#include <iostream>
#include <stdarg.h>

using namespace std;

// initialise variables and such
filePatcher::filePatcher(char *fileName, patch *patchData, int patches, void *pOutputFunction)
{
	// We didn't make any arrays, can't delete them
	bDeleteArrays = false;
	bDeleteNewFileName = false;

	// Assign the default string output function
	determinOutputHandler(pOutputFunction);

	// Open the input file fileName
	_inFile.open(fileName, ios::in|ios::binary);
	if(!_inFile)
	{
		pOutput("%s not found.", fileName);
	}
	else
	{
		int len = (int)strlen(fileName)+1;

		// copy over number of patches to private _patches
		_patches = patches;

		// Create space for the filename and patch array
		_fileName = new char[len];
		_patchData = new patch[patches];

		// Copy the fileName to private _fileName
		strncpy(_fileName, fileName, len);

		// Copy the patch array to private _patchData
		memcpy(_patchData, patchData, patches*sizeof(patch));

		// Start the main patching process
		startupPatcher();

		// We made the array so we need to delete it
		bDeleteArrays = true;
	}
}

// cleanup
filePatcher::~filePatcher()
{
	if(bDeleteArrays)
	{
		delete [] _fileName;
		delete [] _patchData;
	}
	if(bDeleteNewFileName)
	{
		delete [] _newFileName;
	}
	if(_inFile.is_open())
	{
		_inFile.close();
	}
	if(_outFile.is_open())
	{
		_outFile.close();
	}
}

// the main patching process
void filePatcher::startupPatcher()
{
	backupFile();

	if(replaceData(false))
	{
		// if it does, then back up the file
		//backupFile();
		// and then replace the originalData with patchedData
		replaceData(true);
	}
	else
	{
		pOutput("Failed to apply patch");
		return;
	}
	pOutput("Patch successfully applied!\n\nPatched file saved as: %s\n\nMove %s into your Fallout 2 directory and run it instead of Fallout2.exe", _newFileName, _newFileName);
}

// assign an output handler for messages to the user
void filePatcher::determinOutputHandler(void *pOutputFunction)
{
	if(pOutputFunction != NULL)
	{
		pOutput = ((void (*)(char *, ...))pOutputFunction);
	}
	else
	{
		pOutput = &defaultOutputHandler;
	}
}

// Backups the main file, file
void filePatcher::backupFile()
{
	if(!makeBackupFile())
	{
		pOutput("Error creating backup file");
	}
	copyFile();
	_inFile.seekg(0,ios::beg);
	_outFile.close();
}

// Generates a backup file that does not exist already, filename.exe.xxx.exe
bool filePatcher::makeBackupFile()
{
	ifstream fileCheck;
	char *backupFile = new char[strlen(_fileName)+9]; // fileName + .xxx\0
	for(int i = 0; i < 999; i++)
	{
		if(i < 10)
		{
            sprintf(backupFile,"%s.00%i.exe", _fileName, i);
		}
		else if(i < 100)
		{
			sprintf(backupFile,"%s.0%i.exe", _fileName, i);
		}
		else if(i < 1000)
		{
			sprintf(backupFile,"%s.%i.exe", _fileName, i);
		}
		else
		{
			return false;
		}

		fileCheck.open(backupFile);
		if(!fileCheck.is_open())
		{
			_newFileName = new char[strlen(backupFile)+1];
			memcpy(_newFileName, backupFile, strlen(backupFile)+1);
			bDeleteNewFileName = true;

			_outFile.open(backupFile, ios::out|ios::binary);
			return true;
		}
		fileCheck.close();
	}
	return false;
}

// Copies the actual data from file inFile to file OutFile
void filePatcher::copyFile()
{
	if(_inFile.is_open() && _outFile.is_open())
	{
		int ch = 0;
		while((ch = _inFile.get()) != EOF)
		{
			_outFile.put(ch);
		}
	}
	else
	{
		pOutput("Error copying data to new executable");
	}
}

// Parameters:
//   bWriteData 
//    False - just check if the data matches what we want to patch
//    True  - patch the data
//   bPatchOldFile
//    True - Use the patched file already created, used by applyMorePatches()
// Returns:
//   True  - Patch succeeded or patch is ready to be written
//   False - Data did not match what we wanted to patch
bool filePatcher::replaceData(bool bWriteData, bool bPatchOldFile)
{
	if(!bWriteData || bPatchOldFile)
	{
		// close file, clear flags and open the file that was patched by the
		//  startupPatcher()
		_inFile.close();
		_inFile.clear();
		_inFile.open(_newFileName, ios::out|ios::in|ios::binary);
	}

	if(!_inFile)
	{
		pOutput("Error: Failed to find new exe");
		return false;
	}

	for(int i = 0; i < _patches; i++)
	{
		// clear flags and goto offset _patchData[i].offset
		_inFile.clear();
		_inFile.seekg(_patchData[i].offset, ios::beg);

		// if bWriteData, then write the data
		if(bWriteData)
		{
			for(int x = 0; x < _patchData[i].length; x++)
			{
				_inFile.put(_patchData[i].patchedData[x]);
			}
		}
		// if !bWriteData, just compare data from patchData to the actual
		//  file we are patching
		else
		{
			// buffer stores the data we read from the file
			char *buffer = new char[_patchData[i].length];

			// read data from offset _patchData[i].offset
			_inFile.read(buffer, _patchData[i].length);
			// go back to the offset (_inFile.read() caused us to move)
			_inFile.seekg(_patchData[i].offset, ios::beg);

			// if originalData != Buffer, return false
			if(!binncmp(_patchData[i].originalData, buffer, _patchData[i].length))
			{
				pOutput("Patch %i: patch did not match data from file", (i+1));
				return false;
			}

			// buffer is just junk we took from the file stream for processing
			delete [] buffer;
		}
	}

	return true;
}


// Applies more patches to the file. Good for creating a large amount of standard
//  data that won't be changed except for a few bytes and having this function
//  patch the changes
void filePatcher::applyMorePatches(patch *patchData, int patches, bool bSilent)
{
	// open the file we modified from the constructor
	_inFile.clear();
	_inFile.open(_newFileName, ios::out|ios::in|ios::binary);

	// copy patchData to _patchData and patches to _patches
	_patches = patches;
	delete [] _patchData;
	_patchData = new patch[patches];
	memcpy(_patchData, patchData, patches*sizeof(patch));

	if(replaceData(false, true))
	{
		replaceData(true, true);
	}
	else
	{
		pOutput("Failed to apply patch");
		return;
	}

	// Usually good to keep this silent, else it will display this message twice
	if(!bSilent)
	{
		pOutput("Patch successfully applied!\n\nPatched file saved as: %s\n\nMove %s into your Fallout 2 directory and run it instead of Fallout2.exe", _newFileName, _newFileName);
	}
}

// Default string output handler
void filePatcher::defaultOutputHandler(char *buff, ...)
{
	char * str;
	va_list argList;

	va_start(argList, buff);
	str = new char[_vscprintf(buff, argList)];
	vsprintf(str, buff, argList);
	va_end(argList);

#ifdef _WINDOWS_
	MessageBox(0, str, "Default Message Handler", 0);
#else
	cout << str << endl;
#endif
}

// checks each character of every type including null terminator up to len
bool binncmp(char *strA, char *strB, int len)
{
	char *pStrA = strA;
	char *pStrB = strB;

	for(int i = 0; i < len; i++)
	{
		if(*(pStrA++) != *(pStrB++))
		{
			return false;
		}
	}
	return true;
}

// Takes an integer and returns a hex number (string) formatted for assembly
char *numToAsm(unsigned int number)
{
	int size = 9;
	char ch;
	char *tempStr = new char[size];

	sprintf(tempStr, "%.8x", number);
	strrev(tempStr);

	for(int i = 0; i < size-1; i+=2)
	{
		ch = tempStr[i];
		tempStr[i] = tempStr[i+1];
		tempStr[i+1] = ch;
	}

	return tempStr;
}